import click
from . import products
from ..output.table import output_entry
from ..input_helpers import get_org_from_input_or_ctx
from ..orgs import get_raw


@click.command(name="list-products")
@click.option("--limit", default=500)
@click.option("--get-subscription-data", default=True)
@click.pass_context
def cli_command_list_products(ctx, **kwargs):
    results = products.list_products(ctx, **kwargs)
    table = products.format_products(ctx, results)
    print(table)


@click.command(name="add-product")
@click.argument("name")
@click.option("--product-price-id", default=None, multiple=True)
@click.option("--dev-mode", is_flag=True, default=None)
@click.option("--label", default=None)
@click.option("--description", default=None)
@click.option("--trial_period", type=int, default=None)
@click.pass_context
def cli_command_add_product(ctx, product_price_id, **kwargs):
    result = products.add_product(
        ctx, product_price_ids=list(product_price_id), **kwargs
    )
    output_entry(ctx, result.to_dict())


@click.command(name="delete-product")
@click.argument("product-id")
@click.pass_context
def cli_command_delete_product(ctx, **kwargs):
    products.delete_product(ctx, **kwargs)


@click.command(name="show-product")
@click.argument("product_id")
@click.pass_context
def cli_command_show_product(ctx, **kwargs):
    result = products.get_product(ctx, **kwargs)
    output_entry(ctx, result.to_dict())


@click.command(name="show-default-product")
@click.option("--org_id", default=None)
@click.pass_context
def cli_command_show_default_product(ctx, org_id, **kwargs):
    org = get_raw(ctx, get_org_from_input_or_ctx(ctx, org_id))

    print(f"new suborg product-id decision priority for root org: {org.organisation}")
    print("NOTE: priority 0 is highest\n")
    print("priority 0-product_label_override property organisation create")
    feature_override = "not_set"
    for feature_flag in org.feature_flags:
        if feature_flag.feature == "product_label_override_default":
            feature_override = feature_flag.feature

    print(f"priority 1-product_label_override_default={feature_override}")
    print("priority 2-GEO IP for currency product label")


@click.command(name="update-product")
@click.argument("product_id")
@click.option("--dev-mode", is_flag=True, default=None)
@click.option("--name", default=None)
@click.option("--label", default=None)
@click.option("--trial_period", type=int, default=None)
@click.option("--description", default=None)
@click.option("--product-price-id", default=None, multiple=True)
@click.option("--remove-product-price-id", default=None, multiple=True)
@click.option("--feature-id", default=None, multiple=True)
@click.option("--remove-feature-id", default=None, multiple=True)
@click.pass_context
def cli_command_update_product(
    ctx,
    product_price_id,
    remove_product_price_id,
    feature_id,
    remove_feature_id,
    **kwargs,
):
    result = products.update_product(
        ctx,
        remove_product_price_ids=list(remove_product_price_id),
        product_price_ids=list(product_price_id),
        remove_feature_ids=list(remove_feature_id),
        feature_ids=list(feature_id),
        **kwargs,
    )
    output_entry(ctx, result.to_dict())


all_funcs = [func for func in dir() if "cli_command_" in func]


def add_commands(cli):
    glob = globals()
    for func in all_funcs:
        cli.add_command(glob[func])



from unittest.mock import MagicMock

class RulesApiMock:

    def __init__(self):
        self.mock_create_ruleset_label = MagicMock()
        self.mock_create_standalone_rule = MagicMock()
        self.mock_create_standalone_rule_policy = MagicMock()
        self.mock_create_standalone_rule_tree = MagicMock()
        self.mock_create_standalone_ruleset = MagicMock()
        self.mock_create_standalone_ruleset_bundle = MagicMock()
        self.mock_delete_ruleset_label = MagicMock()
        self.mock_delete_standalone_rule = MagicMock()
        self.mock_delete_standalone_rule_policy = MagicMock()
        self.mock_delete_standalone_rule_tree = MagicMock()
        self.mock_delete_standalone_ruleset = MagicMock()
        self.mock_delete_standalone_ruleset_bundle = MagicMock()
        self.mock_get_ruleset_label = MagicMock()
        self.mock_get_standalone_rule = MagicMock()
        self.mock_get_standalone_rule_policy = MagicMock()
        self.mock_get_standalone_rule_tree = MagicMock()
        self.mock_get_standalone_ruleset = MagicMock()
        self.mock_get_standalone_ruleset_bundle = MagicMock()
        self.mock_list_ruleset_labels = MagicMock()
        self.mock_list_standalone_rule_policies = MagicMock()
        self.mock_list_standalone_rule_trees = MagicMock()
        self.mock_list_standalone_rules = MagicMock()
        self.mock_list_standalone_ruleset_bundles = MagicMock()
        self.mock_list_standalone_rulesets = MagicMock()
        self.mock_replace_standalone_rule = MagicMock()
        self.mock_replace_standalone_rule_policy = MagicMock()
        self.mock_replace_standalone_rule_tree = MagicMock()
        self.mock_replace_standalone_ruleset = MagicMock()
        self.mock_replace_standalone_ruleset_bundle = MagicMock()

    def create_ruleset_label(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.create_ruleset_label with MagicMock.
        """
        return self.mock_create_ruleset_label(self, *args, **kwargs)

    def create_standalone_rule(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.create_standalone_rule with MagicMock.
        """
        return self.mock_create_standalone_rule(self, *args, **kwargs)

    def create_standalone_rule_policy(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.create_standalone_rule_policy with MagicMock.
        """
        return self.mock_create_standalone_rule_policy(self, *args, **kwargs)

    def create_standalone_rule_tree(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.create_standalone_rule_tree with MagicMock.
        """
        return self.mock_create_standalone_rule_tree(self, *args, **kwargs)

    def create_standalone_ruleset(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.create_standalone_ruleset with MagicMock.
        """
        return self.mock_create_standalone_ruleset(self, *args, **kwargs)

    def create_standalone_ruleset_bundle(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.create_standalone_ruleset_bundle with MagicMock.
        """
        return self.mock_create_standalone_ruleset_bundle(self, *args, **kwargs)

    def delete_ruleset_label(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.delete_ruleset_label with MagicMock.
        """
        return self.mock_delete_ruleset_label(self, *args, **kwargs)

    def delete_standalone_rule(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.delete_standalone_rule with MagicMock.
        """
        return self.mock_delete_standalone_rule(self, *args, **kwargs)

    def delete_standalone_rule_policy(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.delete_standalone_rule_policy with MagicMock.
        """
        return self.mock_delete_standalone_rule_policy(self, *args, **kwargs)

    def delete_standalone_rule_tree(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.delete_standalone_rule_tree with MagicMock.
        """
        return self.mock_delete_standalone_rule_tree(self, *args, **kwargs)

    def delete_standalone_ruleset(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.delete_standalone_ruleset with MagicMock.
        """
        return self.mock_delete_standalone_ruleset(self, *args, **kwargs)

    def delete_standalone_ruleset_bundle(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.delete_standalone_ruleset_bundle with MagicMock.
        """
        return self.mock_delete_standalone_ruleset_bundle(self, *args, **kwargs)

    def get_ruleset_label(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.get_ruleset_label with MagicMock.
        """
        return self.mock_get_ruleset_label(self, *args, **kwargs)

    def get_standalone_rule(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.get_standalone_rule with MagicMock.
        """
        return self.mock_get_standalone_rule(self, *args, **kwargs)

    def get_standalone_rule_policy(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.get_standalone_rule_policy with MagicMock.
        """
        return self.mock_get_standalone_rule_policy(self, *args, **kwargs)

    def get_standalone_rule_tree(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.get_standalone_rule_tree with MagicMock.
        """
        return self.mock_get_standalone_rule_tree(self, *args, **kwargs)

    def get_standalone_ruleset(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.get_standalone_ruleset with MagicMock.
        """
        return self.mock_get_standalone_ruleset(self, *args, **kwargs)

    def get_standalone_ruleset_bundle(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.get_standalone_ruleset_bundle with MagicMock.
        """
        return self.mock_get_standalone_ruleset_bundle(self, *args, **kwargs)

    def list_ruleset_labels(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.list_ruleset_labels with MagicMock.
        """
        return self.mock_list_ruleset_labels(self, *args, **kwargs)

    def list_standalone_rule_policies(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.list_standalone_rule_policies with MagicMock.
        """
        return self.mock_list_standalone_rule_policies(self, *args, **kwargs)

    def list_standalone_rule_trees(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.list_standalone_rule_trees with MagicMock.
        """
        return self.mock_list_standalone_rule_trees(self, *args, **kwargs)

    def list_standalone_rules(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.list_standalone_rules with MagicMock.
        """
        return self.mock_list_standalone_rules(self, *args, **kwargs)

    def list_standalone_ruleset_bundles(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.list_standalone_ruleset_bundles with MagicMock.
        """
        return self.mock_list_standalone_ruleset_bundles(self, *args, **kwargs)

    def list_standalone_rulesets(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.list_standalone_rulesets with MagicMock.
        """
        return self.mock_list_standalone_rulesets(self, *args, **kwargs)

    def replace_standalone_rule(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.replace_standalone_rule with MagicMock.
        """
        return self.mock_replace_standalone_rule(self, *args, **kwargs)

    def replace_standalone_rule_policy(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.replace_standalone_rule_policy with MagicMock.
        """
        return self.mock_replace_standalone_rule_policy(self, *args, **kwargs)

    def replace_standalone_rule_tree(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.replace_standalone_rule_tree with MagicMock.
        """
        return self.mock_replace_standalone_rule_tree(self, *args, **kwargs)

    def replace_standalone_ruleset(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.replace_standalone_ruleset with MagicMock.
        """
        return self.mock_replace_standalone_ruleset(self, *args, **kwargs)

    def replace_standalone_ruleset_bundle(self, *args, **kwargs):
        """
        This method mocks the original api RulesApi.replace_standalone_ruleset_bundle with MagicMock.
        """
        return self.mock_replace_standalone_ruleset_bundle(self, *args, **kwargs)




"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2024.05.09
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.model_utils import (  # noqa: F401
    ApiTypeError,
    ModelComposed,
    ModelNormal,
    ModelSimple,
    cached_property,
    change_keys_js_to_python,
    convert_js_args_to_python_args,
    date,
    datetime,
    file_type,
    none_type,
    validate_get_composed_info,
)
from ..model_utils import OpenApiModel
from agilicus_api.exceptions import ApiAttributeError


def lazy_import():
    from agilicus_api.model.alternate_mode_setting import AlternateModeSetting
    from agilicus_api.model.application_assignment import ApplicationAssignment
    from agilicus_api.model.application_monitoring_config import ApplicationMonitoringConfig
    from agilicus_api.model.application_state_selector import ApplicationStateSelector
    from agilicus_api.model.definition import Definition
    from agilicus_api.model.display_info import DisplayInfo
    from agilicus_api.model.environment import Environment
    from agilicus_api.model.k8s_slug import K8sSlug
    from agilicus_api.model.role_list import RoleList
    from agilicus_api.model.roles_config import RolesConfig
    from agilicus_api.model.rules_config import RulesConfig
    from agilicus_api.model.workload_configuration import WorkloadConfiguration
    globals()['AlternateModeSetting'] = AlternateModeSetting
    globals()['ApplicationAssignment'] = ApplicationAssignment
    globals()['ApplicationMonitoringConfig'] = ApplicationMonitoringConfig
    globals()['ApplicationStateSelector'] = ApplicationStateSelector
    globals()['Definition'] = Definition
    globals()['DisplayInfo'] = DisplayInfo
    globals()['Environment'] = Environment
    globals()['K8sSlug'] = K8sSlug
    globals()['RoleList'] = RoleList
    globals()['RolesConfig'] = RolesConfig
    globals()['RulesConfig'] = RulesConfig
    globals()['WorkloadConfiguration'] = WorkloadConfiguration


class Application(ModelNormal):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Attributes:
      allowed_values (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          with a capitalized key describing the allowed value and an allowed
          value. These dicts store the allowed enum values.
      attribute_map (dict): The key is attribute name
          and the value is json key in definition.
      discriminator_value_class_map (dict): A dict to go from the discriminator
          variable value to the discriminator class name.
      validations (dict): The key is the tuple path to the attribute
          and the for var_name this is (var_name,). The value is a dict
          that stores validations for max_length, min_length, max_items,
          min_items, exclusive_maximum, inclusive_maximum, exclusive_minimum,
          inclusive_minimum, and regex.
      additional_properties_type (tuple): A tuple of classes accepted
          as additional properties values.
    """

    allowed_values = {
        ('image_credentials_type',): {
            'BASIC_AUTH': "basic_auth",
            'TOKEN': "token",
        },
        ('published',): {
            'NO': "no",
            'PUBLIC': "public",
        },
        ('location',): {
            'HOSTED': "hosted",
            'EXTERNAL': "external",
        },
        ('application_type',): {
            'USER_DEFINED': "user_defined",
            'INTERNAL': "internal",
        },
    }

    validations = {
        ('name',): {
            'max_length': 63,
            'min_length': 1,
            'regex': {
                'pattern': r'^[a-zA-Z0-9-:]+$',  # noqa: E501
            },
        },
        ('category',): {
            'max_length': 100,
            'min_length': 1,
        },
        ('org_id',): {
            'max_length': 40,
            'min_length': 1,
        },
        ('image_username',): {
            'max_length': 100,
            'min_length': 1,
        },
        ('image_password',): {
            'max_length': 100,
            'min_length': 1,
        },
        ('contact_email',): {
            'max_length': 100,
        },
        ('default_role_id',): {
            'max_length': 40,
        },
        ('default_role_name',): {
            'max_length': 100,
            'min_length': 1,
        },
        ('icon_url',): {
            'max_length': 1024,
        },
        ('service_account_id',): {
            'max_length': 40,
            'min_length': 1,
        },
    }

    @property
    def created(self):
       return self.get("created")

    @created.setter
    def created(self, new_value):
       self.created = new_value

    @property
    def id(self):
       return self.get("id")

    @id.setter
    def id(self, new_value):
       self.id = new_value

    @property
    def parent_id(self):
       return self.get("parent_id")

    @parent_id.setter
    def parent_id(self, new_value):
       self.parent_id = new_value

    @property
    def name(self):
       return self.get("name")

    @name.setter
    def name(self, new_value):
       self.name = new_value

    @property
    def description(self):
       return self.get("description")

    @description.setter
    def description(self, new_value):
       self.description = new_value

    @property
    def category(self):
       return self.get("category")

    @category.setter
    def category(self, new_value):
       self.category = new_value

    @property
    def image(self):
       return self.get("image")

    @image.setter
    def image(self, new_value):
       self.image = new_value

    @property
    def image_username(self):
       return self.get("image_username")

    @image_username.setter
    def image_username(self, new_value):
       self.image_username = new_value

    @property
    def image_password(self):
       return self.get("image_password")

    @image_password.setter
    def image_password(self, new_value):
       self.image_password = new_value

    @property
    def image_credentials_type(self):
       return self.get("image_credentials_type")

    @image_credentials_type.setter
    def image_credentials_type(self, new_value):
       self.image_credentials_type = new_value

    @property
    def environments(self):
       return self.get("environments")

    @environments.setter
    def environments(self, new_value):
       self.environments = new_value

    @property
    def org_id(self):
       return self.get("org_id")

    @org_id.setter
    def org_id(self, new_value):
       self.org_id = new_value

    @property
    def contact_email(self):
       return self.get("contact_email")

    @contact_email.setter
    def contact_email(self, new_value):
       self.contact_email = new_value

    @property
    def monitoring_config(self):
       return self.get("monitoring_config")

    @monitoring_config.setter
    def monitoring_config(self, new_value):
       self.monitoring_config = new_value

    @property
    def port(self):
       return self.get("port")

    @port.setter
    def port(self, new_value):
       self.port = new_value

    @property
    def healthcheck_uri(self):
       return self.get("healthcheck_uri")

    @healthcheck_uri.setter
    def healthcheck_uri(self, new_value):
       self.healthcheck_uri = new_value

    @property
    def roles(self):
       return self.get("roles")

    @roles.setter
    def roles(self, new_value):
       self.roles = new_value

    @property
    def roles_config(self):
       return self.get("roles_config")

    @roles_config.setter
    def roles_config(self, new_value):
       self.roles_config = new_value

    @property
    def rules_config(self):
       return self.get("rules_config")

    @rules_config.setter
    def rules_config(self, new_value):
       self.rules_config = new_value

    @property
    def definitions(self):
       return self.get("definitions")

    @definitions.setter
    def definitions(self, new_value):
       self.definitions = new_value

    @property
    def assignments(self):
       return self.get("assignments")

    @assignments.setter
    def assignments(self, new_value):
       self.assignments = new_value

    @property
    def owned(self):
       return self.get("owned")

    @owned.setter
    def owned(self, new_value):
       self.owned = new_value

    @property
    def maintained(self):
       return self.get("maintained")

    @maintained.setter
    def maintained(self, new_value):
       self.maintained = new_value

    @property
    def assigned(self):
       return self.get("assigned")

    @assigned.setter
    def assigned(self, new_value):
       self.assigned = new_value

    @property
    def published(self):
       return self.get("published")

    @published.setter
    def published(self, new_value):
       self.published = new_value

    @property
    def default_role_id(self):
       return self.get("default_role_id")

    @default_role_id.setter
    def default_role_id(self, new_value):
       self.default_role_id = new_value

    @property
    def default_role_name(self):
       return self.get("default_role_name")

    @default_role_name.setter
    def default_role_name(self, new_value):
       self.default_role_name = new_value

    @property
    def icon_url(self):
       return self.get("icon_url")

    @icon_url.setter
    def icon_url(self, new_value):
       self.icon_url = new_value

    @property
    def display_info(self):
       return self.get("display_info")

    @display_info.setter
    def display_info(self, new_value):
       self.display_info = new_value

    @property
    def updated(self):
       return self.get("updated")

    @updated.setter
    def updated(self, new_value):
       self.updated = new_value

    @property
    def location(self):
       return self.get("location")

    @location.setter
    def location(self, new_value):
       self.location = new_value

    @property
    def service_account_id(self):
       return self.get("service_account_id")

    @service_account_id.setter
    def service_account_id(self, new_value):
       self.service_account_id = new_value

    @property
    def service_account_required(self):
       return self.get("service_account_required")

    @service_account_required.setter
    def service_account_required(self, new_value):
       self.service_account_required = new_value

    @property
    def application_type(self):
       return self.get("application_type")

    @application_type.setter
    def application_type(self, new_value):
       self.application_type = new_value

    @property
    def name_slug(self):
       return self.get("name_slug")

    @name_slug.setter
    def name_slug(self, new_value):
       self.name_slug = new_value

    @property
    def workload_config(self):
       return self.get("workload_config")

    @workload_config.setter
    def workload_config(self, new_value):
       self.workload_config = new_value

    @property
    def admin_state(self):
       return self.get("admin_state")

    @admin_state.setter
    def admin_state(self, new_value):
       self.admin_state = new_value

    @property
    def alternate_mode_setting(self):
       return self.get("alternate_mode_setting")

    @alternate_mode_setting.setter
    def alternate_mode_setting(self, new_value):
       self.alternate_mode_setting = new_value

    @property
    def launchers(self):
       return self.get("launchers")

    @launchers.setter
    def launchers(self, new_value):
       self.launchers = new_value

    @cached_property
    def additional_properties_type():
        """
        This must be a method because a model may have properties that are
        of type self, this must run after the class is loaded
        """
        lazy_import()
        return (bool, date, datetime, dict, float, int, list, str, none_type,)  # noqa: E501

    _nullable = False

    @cached_property
    def openapi_types():
        """
        This must be a method because a model may have properties that are
        of type self, this must run after the class is loaded

        Returns
            openapi_types (dict): The key is attribute name
                and the value is attribute type.
        """
        lazy_import()
        return {
            'name': (str,),  # noqa: E501
            'category': (str,),  # noqa: E501
            'org_id': (str,),  # noqa: E501
            'created': (datetime,),  # noqa: E501
            'id': (str,),  # noqa: E501
            'parent_id': (str,),  # noqa: E501
            'description': (str,),  # noqa: E501
            'image': (str,),  # noqa: E501
            'image_username': (str,),  # noqa: E501
            'image_password': (str,),  # noqa: E501
            'image_credentials_type': (str,),  # noqa: E501
            'environments': ([Environment],),  # noqa: E501
            'contact_email': (str,),  # noqa: E501
            'monitoring_config': (ApplicationMonitoringConfig,),  # noqa: E501
            'port': (int,),  # noqa: E501
            'healthcheck_uri': (str,),  # noqa: E501
            'roles': (RoleList,),  # noqa: E501
            'roles_config': (RolesConfig,),  # noqa: E501
            'rules_config': (RulesConfig,),  # noqa: E501
            'definitions': ([Definition],),  # noqa: E501
            'assignments': ([ApplicationAssignment],),  # noqa: E501
            'owned': (bool,),  # noqa: E501
            'maintained': (bool,),  # noqa: E501
            'assigned': (bool,),  # noqa: E501
            'published': (str,),  # noqa: E501
            'default_role_id': (str, none_type,),  # noqa: E501
            'default_role_name': (str, none_type,),  # noqa: E501
            'icon_url': (str,),  # noqa: E501
            'display_info': (DisplayInfo,),  # noqa: E501
            'updated': (datetime,),  # noqa: E501
            'location': (str,),  # noqa: E501
            'service_account_id': (str,),  # noqa: E501
            'service_account_required': (bool,),  # noqa: E501
            'application_type': (str,),  # noqa: E501
            'name_slug': (K8sSlug,),  # noqa: E501
            'workload_config': (WorkloadConfiguration,),  # noqa: E501
            'admin_state': (ApplicationStateSelector,),  # noqa: E501
            'alternate_mode_setting': (AlternateModeSetting,),  # noqa: E501
            'launchers': ([str],),  # noqa: E501
        }

    @cached_property
    def discriminator():
        return None



    attribute_map = {
        'name': 'name',  # noqa: E501
        'category': 'category',  # noqa: E501
        'org_id': 'org_id',  # noqa: E501
        'created': 'created',  # noqa: E501
        'id': 'id',  # noqa: E501
        'parent_id': 'parent_id',  # noqa: E501
        'description': 'description',  # noqa: E501
        'image': 'image',  # noqa: E501
        'image_username': 'image_username',  # noqa: E501
        'image_password': 'image_password',  # noqa: E501
        'image_credentials_type': 'image_credentials_type',  # noqa: E501
        'environments': 'environments',  # noqa: E501
        'contact_email': 'contact_email',  # noqa: E501
        'monitoring_config': 'monitoring_config',  # noqa: E501
        'port': 'port',  # noqa: E501
        'healthcheck_uri': 'healthcheck_uri',  # noqa: E501
        'roles': 'roles',  # noqa: E501
        'roles_config': 'roles_config',  # noqa: E501
        'rules_config': 'rules_config',  # noqa: E501
        'definitions': 'definitions',  # noqa: E501
        'assignments': 'assignments',  # noqa: E501
        'owned': 'owned',  # noqa: E501
        'maintained': 'maintained',  # noqa: E501
        'assigned': 'assigned',  # noqa: E501
        'published': 'published',  # noqa: E501
        'default_role_id': 'default_role_id',  # noqa: E501
        'default_role_name': 'default_role_name',  # noqa: E501
        'icon_url': 'icon_url',  # noqa: E501
        'display_info': 'display_info',  # noqa: E501
        'updated': 'updated',  # noqa: E501
        'location': 'location',  # noqa: E501
        'service_account_id': 'service_account_id',  # noqa: E501
        'service_account_required': 'service_account_required',  # noqa: E501
        'application_type': 'application_type',  # noqa: E501
        'name_slug': 'name_slug',  # noqa: E501
        'workload_config': 'workload_config',  # noqa: E501
        'admin_state': 'admin_state',  # noqa: E501
        'alternate_mode_setting': 'alternate_mode_setting',  # noqa: E501
        'launchers': 'launchers',  # noqa: E501
    }

    read_only_vars = {
        'created',  # noqa: E501
        'id',  # noqa: E501
        'parent_id',  # noqa: E501
        'default_role_name',  # noqa: E501
        'updated',  # noqa: E501
        'service_account_id',  # noqa: E501
    }

    _composed_schemas = {}

    @classmethod
    @convert_js_args_to_python_args
    def _from_openapi_data(cls, name, category, org_id, *args, **kwargs):  # noqa: E501
        """Application - a model defined in OpenAPI

        Args:
            name (str): application name
            category (str): application category
            org_id (str): organisation id

        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _spec_property_naming (bool): True if the variable names in the input data
                                are serialized names, as specified in the OpenAPI document.
                                False if the variable names in the input data
                                are pythonic names, e.g. snake case (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            _visited_composed_classes (tuple): This stores a tuple of
                                classes that we have traveled through so that
                                if we see that class again we will not use its
                                discriminator again.
                                When traveling through a discriminator, the
                                composed schema that is
                                is traveled through is added to this set.
                                For example if Animal has a discriminator
                                petType and we pass in "Dog", and the class Dog
                                allOf includes Animal, we move through Animal
                                once using the discriminator, and pick Dog.
                                Then in Dog, we will make an instance of the
                                Animal class but this time we won't travel
                                through its discriminator because we passed in
                                _visited_composed_classes = (Animal,)
            created (datetime): Creation time. [optional]  # noqa: E501
            id (str): application service id. [optional]  # noqa: E501
            parent_id (str): For inheritance, defines the parent id.. [optional]  # noqa: E501
            description (str): Application description text. [optional]  # noqa: E501
            image (str): image registry path and name. [optional]  # noqa: E501
            image_username (str): registry username. [optional]  # noqa: E501
            image_password (str): registry password. [optional]  # noqa: E501
            image_credentials_type (str): type of credentials for image. [optional]  # noqa: E501
            environments ([Environment]): List of environments. [optional]  # noqa: E501
            contact_email (str): Administrator contact email. [optional]  # noqa: E501
            monitoring_config (ApplicationMonitoringConfig): [optional]  # noqa: E501
            port (int): The transport layer port the application listens on for requests. E.g. if the application listens for HTTP requests on the standard port, port 80, set this to 80. . [optional]  # noqa: E501
            healthcheck_uri (str): health check URI. [optional]  # noqa: E501
            roles (RoleList): [optional]  # noqa: E501
            roles_config (RolesConfig): [optional]  # noqa: E501
            rules_config (RulesConfig): [optional]  # noqa: E501
            definitions ([Definition]): List of definitions. [optional]  # noqa: E501
            assignments ([ApplicationAssignment]): Controls the Organisations which have access to Environments of this Application. . [optional]  # noqa: E501
            owned (bool): Whether this Application is owned by the provided organisation. . [optional]  # noqa: E501
            maintained (bool): Whether this Application has an Environment maintained by the provided organisation. . [optional]  # noqa: E501
            assigned (bool): Whether an Environment is assigned to this Application. . [optional]  # noqa: E501
            published (str): Whether or not this Application is published, and if so, how. An application that has been published somewhere will have high level details about it visible (through apis intended to expose published applications), such as its name and description. The enum values mean the following:  - no: This application is not published. It will only be visibile to users with       permission to access the application, or to administrators.  - public: This application is published to the public catalogue. Any user who       can request access to the organisation will see high level details about this       application. . [optional] if omitted the server will use the default value of "no"  # noqa: E501
            default_role_id (str, none_type): The id of the default role to use for this application. This role will be assigned to users granted access to this application by default, unless overridden. This value must point to a valid Role for this application. A string of length 0 will clear out the default role.. [optional]  # noqa: E501
            default_role_name (str, none_type): The name of the default role. Does not change which role is used. Provided for convenience. . [optional]  # noqa: E501
            icon_url (str): A url pointing to an icon representing this application. . [optional]  # noqa: E501
            display_info (DisplayInfo): [optional]  # noqa: E501
            updated (datetime): Update time. [optional]  # noqa: E501
            location (str): Where the application runs. Applications can run within an Agilicus hosting platform, or elsewhere on the internet. The possible values have the following meanings:   - `hosted`: the application runs on the Agilicus hosting platform.   - `external`: the application runs elsewhere on the internet. External applications     will not create instances in the Agilicus hosting platform. However, they will     be able to use Agilicus for Identity and Authentication. . [optional]  # noqa: E501
            service_account_id (str): Service account user GUID used to deploy the application. [optional]  # noqa: E501
            service_account_required (bool): If service_account_enabled field is set to true, a service account will be created. If service_account_enabled field is set to false, the service account will be deleted. If the service_account_enabled field is not set no action on the service account is taken. . [optional]  # noqa: E501
            application_type (str): The type of application meanings:   - `user_defined`: A user defined application.   - `internal`: An internal application used by agilicus services. . [optional] if omitted the server will use the default value of "user_defined"  # noqa: E501
            name_slug (K8sSlug): [optional]  # noqa: E501
            workload_config (WorkloadConfiguration): [optional]  # noqa: E501
            admin_state (ApplicationStateSelector): [optional]  # noqa: E501
            alternate_mode_setting (AlternateModeSetting): [optional]  # noqa: E501
            launchers ([str]): A list of launchers that can be used to launch this application. . [optional]  # noqa: E501
        """

        _check_type = kwargs.pop('_check_type', True)
        _spec_property_naming = kwargs.pop('_spec_property_naming', False)
        _path_to_item = kwargs.pop('_path_to_item', ())
        _configuration = kwargs.pop('_configuration', None)
        _visited_composed_classes = kwargs.pop('_visited_composed_classes', ())

        self = super(OpenApiModel, cls).__new__(cls)

        if args:
            raise ApiTypeError(
                "Invalid positional arguments=%s passed to %s. Remove those invalid positional arguments." % (
                    args,
                    self.__class__.__name__,
                ),
                path_to_item=_path_to_item,
                valid_classes=(self.__class__,),
            )

        self._data_store = {}
        self._check_type = _check_type
        self._spec_property_naming = _spec_property_naming
        self._path_to_item = _path_to_item
        self._configuration = _configuration
        self._visited_composed_classes = _visited_composed_classes + (self.__class__,)

        self.name = name
        self.category = category
        self.org_id = org_id
        for var_name, var_value in kwargs.items():
            if var_name not in self.attribute_map and \
                        self._configuration is not None and \
                        self._configuration.discard_unknown_keys and \
                        self.additional_properties_type is None:
                # discard variable.
                continue
            setattr(self, var_name, var_value)
        return self

    def __python_set(val):
        return set(val)
 
    required_properties = __python_set([
        '_data_store',
        '_check_type',
        '_spec_property_naming',
        '_path_to_item',
        '_configuration',
        '_visited_composed_classes',
    ])

    @convert_js_args_to_python_args
    def __init__(self, name, category, org_id, *args, **kwargs):  # noqa: E501
        """Application - a model defined in OpenAPI

        Args:
            name (str): application name
            category (str): application category
            org_id (str): organisation id

        Keyword Args:
            _check_type (bool): if True, values for parameters in openapi_types
                                will be type checked and a TypeError will be
                                raised if the wrong type is input.
                                Defaults to True
            _path_to_item (tuple/list): This is a list of keys or values to
                                drill down to the model in received_data
                                when deserializing a response
            _spec_property_naming (bool): True if the variable names in the input data
                                are serialized names, as specified in the OpenAPI document.
                                False if the variable names in the input data
                                are pythonic names, e.g. snake case (default)
            _configuration (Configuration): the instance to use when
                                deserializing a file_type parameter.
                                If passed, type conversion is attempted
                                If omitted no type conversion is done.
            _visited_composed_classes (tuple): This stores a tuple of
                                classes that we have traveled through so that
                                if we see that class again we will not use its
                                discriminator again.
                                When traveling through a discriminator, the
                                composed schema that is
                                is traveled through is added to this set.
                                For example if Animal has a discriminator
                                petType and we pass in "Dog", and the class Dog
                                allOf includes Animal, we move through Animal
                                once using the discriminator, and pick Dog.
                                Then in Dog, we will make an instance of the
                                Animal class but this time we won't travel
                                through its discriminator because we passed in
                                _visited_composed_classes = (Animal,)
            created (datetime): Creation time. [optional]  # noqa: E501
            id (str): application service id. [optional]  # noqa: E501
            parent_id (str): For inheritance, defines the parent id.. [optional]  # noqa: E501
            description (str): Application description text. [optional]  # noqa: E501
            image (str): image registry path and name. [optional]  # noqa: E501
            image_username (str): registry username. [optional]  # noqa: E501
            image_password (str): registry password. [optional]  # noqa: E501
            image_credentials_type (str): type of credentials for image. [optional]  # noqa: E501
            environments ([Environment]): List of environments. [optional]  # noqa: E501
            contact_email (str): Administrator contact email. [optional]  # noqa: E501
            monitoring_config (ApplicationMonitoringConfig): [optional]  # noqa: E501
            port (int): The transport layer port the application listens on for requests. E.g. if the application listens for HTTP requests on the standard port, port 80, set this to 80. . [optional]  # noqa: E501
            healthcheck_uri (str): health check URI. [optional]  # noqa: E501
            roles (RoleList): [optional]  # noqa: E501
            roles_config (RolesConfig): [optional]  # noqa: E501
            rules_config (RulesConfig): [optional]  # noqa: E501
            definitions ([Definition]): List of definitions. [optional]  # noqa: E501
            assignments ([ApplicationAssignment]): Controls the Organisations which have access to Environments of this Application. . [optional]  # noqa: E501
            owned (bool): Whether this Application is owned by the provided organisation. . [optional]  # noqa: E501
            maintained (bool): Whether this Application has an Environment maintained by the provided organisation. . [optional]  # noqa: E501
            assigned (bool): Whether an Environment is assigned to this Application. . [optional]  # noqa: E501
            published (str): Whether or not this Application is published, and if so, how. An application that has been published somewhere will have high level details about it visible (through apis intended to expose published applications), such as its name and description. The enum values mean the following:  - no: This application is not published. It will only be visibile to users with       permission to access the application, or to administrators.  - public: This application is published to the public catalogue. Any user who       can request access to the organisation will see high level details about this       application. . [optional] if omitted the server will use the default value of "no"  # noqa: E501
            default_role_id (str, none_type): The id of the default role to use for this application. This role will be assigned to users granted access to this application by default, unless overridden. This value must point to a valid Role for this application. A string of length 0 will clear out the default role.. [optional]  # noqa: E501
            default_role_name (str, none_type): The name of the default role. Does not change which role is used. Provided for convenience. . [optional]  # noqa: E501
            icon_url (str): A url pointing to an icon representing this application. . [optional]  # noqa: E501
            display_info (DisplayInfo): [optional]  # noqa: E501
            updated (datetime): Update time. [optional]  # noqa: E501
            location (str): Where the application runs. Applications can run within an Agilicus hosting platform, or elsewhere on the internet. The possible values have the following meanings:   - `hosted`: the application runs on the Agilicus hosting platform.   - `external`: the application runs elsewhere on the internet. External applications     will not create instances in the Agilicus hosting platform. However, they will     be able to use Agilicus for Identity and Authentication. . [optional]  # noqa: E501
            service_account_id (str): Service account user GUID used to deploy the application. [optional]  # noqa: E501
            service_account_required (bool): If service_account_enabled field is set to true, a service account will be created. If service_account_enabled field is set to false, the service account will be deleted. If the service_account_enabled field is not set no action on the service account is taken. . [optional]  # noqa: E501
            application_type (str): The type of application meanings:   - `user_defined`: A user defined application.   - `internal`: An internal application used by agilicus services. . [optional] if omitted the server will use the default value of "user_defined"  # noqa: E501
            name_slug (K8sSlug): [optional]  # noqa: E501
            workload_config (WorkloadConfiguration): [optional]  # noqa: E501
            admin_state (ApplicationStateSelector): [optional]  # noqa: E501
            alternate_mode_setting (AlternateModeSetting): [optional]  # noqa: E501
            launchers ([str]): A list of launchers that can be used to launch this application. . [optional]  # noqa: E501
        """

        _check_type = kwargs.pop('_check_type', True)
        _spec_property_naming = kwargs.pop('_spec_property_naming', False)
        _path_to_item = kwargs.pop('_path_to_item', ())
        _configuration = kwargs.pop('_configuration', None)
        _visited_composed_classes = kwargs.pop('_visited_composed_classes', ())

        if args:
            raise ApiTypeError(
                "Invalid positional arguments=%s passed to %s. Remove those invalid positional arguments." % (
                    args,
                    self.__class__.__name__,
                ),
                path_to_item=_path_to_item,
                valid_classes=(self.__class__,),
            )

        self._data_store = {}
        self._check_type = _check_type
        self._spec_property_naming = _spec_property_naming
        self._path_to_item = _path_to_item
        self._configuration = _configuration
        self._visited_composed_classes = _visited_composed_classes + (self.__class__,)

        self.name = name
        self.category = category
        self.org_id = org_id
        for var_name, var_value in kwargs.items():
            if var_name not in self.attribute_map and \
                        self._configuration is not None and \
                        self._configuration.discard_unknown_keys and \
                        self.additional_properties_type is None:
                # discard variable.
                continue
            setattr(self, var_name, var_value)
            if var_name in self.read_only_vars:
                raise ApiAttributeError(f"`{var_name}` is a read-only attribute. Use `from_openapi_data` to instantiate "
                                     f"class with read only attributes.")




"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2024.05.09
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.api_client import ApiClient, Endpoint as _Endpoint
from agilicus_api.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from agilicus_api.model.add_group_member_request import AddGroupMemberRequest
from agilicus_api.model.email import Email
from agilicus_api.model.error_message import ErrorMessage
from agilicus_api.model.group import Group
from agilicus_api.model.list_groups_response import ListGroupsResponse
from agilicus_api.model.upstream_group_reconcile import UpstreamGroupReconcile
from agilicus_api.model.upstream_group_reconcile_response import UpstreamGroupReconcileResponse
from agilicus_api.model.user import User


class GroupsApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __add_group_member(
            self,
            group_id,
            add_group_member_request,
            **kwargs
        ):
            """Add a group member  # noqa: E501

            Add a group member  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.add_group_member(group_id, add_group_member_request, async_req=True)
            >>> result = thread.get()

            Args:
                group_id (str): group_id path
                add_group_member_request (AddGroupMemberRequest):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                User
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['group_id'] = \
                group_id
            kwargs['add_group_member_request'] = \
                add_group_member_request
            return self.call_with_http_info(**kwargs)

        if self.add_group_member is None:
            self.add_group_member = _Endpoint(
                settings={
                    'response_type': (User,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups/{group_id}/members',
                    'operation_id': 'add_group_member',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'group_id',
                        'add_group_member_request',
                    ],
                    'required': [
                        'group_id',
                        'add_group_member_request',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'group_id':
                            (str,),
                        'add_group_member_request':
                            (AddGroupMemberRequest,),
                    },
                    'attribute_map': {
                        'group_id': 'group_id',
                    },
                    'location_map': {
                        'group_id': 'path',
                        'add_group_member_request': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__add_group_member
            )

        def __create_group(
            self,
            group,
            **kwargs
        ):
            """Create a group  # noqa: E501

            Create a group  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_group(group, async_req=True)
            >>> result = thread.get()

            Args:
                group (Group):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Group
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['group'] = \
                group
            return self.call_with_http_info(**kwargs)

        if self.create_group is None:
            self.create_group = _Endpoint(
                settings={
                    'response_type': (Group,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups',
                    'operation_id': 'create_group',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'group',
                    ],
                    'required': [
                        'group',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'group':
                            (Group,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'group': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_group
            )

        def __create_upstream_group_reconcile(
            self,
            upstream_group_reconcile,
            **kwargs
        ):
            """Reconcile a user's groups based on the user's upstream group membership  # noqa: E501

            Reconcile a user's groups based on the user's upstream group membership  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_upstream_group_reconcile(upstream_group_reconcile, async_req=True)
            >>> result = thread.get()

            Args:
                upstream_group_reconcile (UpstreamGroupReconcile):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                UpstreamGroupReconcileResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['upstream_group_reconcile'] = \
                upstream_group_reconcile
            return self.call_with_http_info(**kwargs)

        if self.create_upstream_group_reconcile is None:
            self.create_upstream_group_reconcile = _Endpoint(
                settings={
                    'response_type': (UpstreamGroupReconcileResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups/reconcile_upstream_groups',
                    'operation_id': 'create_upstream_group_reconcile',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'upstream_group_reconcile',
                    ],
                    'required': [
                        'upstream_group_reconcile',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'upstream_group_reconcile':
                            (UpstreamGroupReconcile,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'upstream_group_reconcile': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_upstream_group_reconcile
            )

        def __create_upstream_group_reconcile_sim(
            self,
            upstream_group_reconcile,
            **kwargs
        ):
            """Get the set of groups to reconcile based on the user's upstream group membership  # noqa: E501

            Get the set of groups to reconcile based on the user's upstream group membership This indicates what reconciliatory action will be taken if this endpoint is invoked via a POST.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_upstream_group_reconcile_sim(upstream_group_reconcile, async_req=True)
            >>> result = thread.get()

            Args:
                upstream_group_reconcile (UpstreamGroupReconcile):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                UpstreamGroupReconcileResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['upstream_group_reconcile'] = \
                upstream_group_reconcile
            return self.call_with_http_info(**kwargs)

        if self.create_upstream_group_reconcile_sim is None:
            self.create_upstream_group_reconcile_sim = _Endpoint(
                settings={
                    'response_type': (UpstreamGroupReconcileResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups/simulate_reconcile_upstream_groups',
                    'operation_id': 'create_upstream_group_reconcile_sim',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'upstream_group_reconcile',
                    ],
                    'required': [
                        'upstream_group_reconcile',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'upstream_group_reconcile':
                            (UpstreamGroupReconcile,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'upstream_group_reconcile': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_upstream_group_reconcile_sim
            )

        def __delete_group(
            self,
            group_id,
            **kwargs
        ):
            """Delete a group  # noqa: E501

            Delete a group  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_group(group_id, async_req=True)
            >>> result = thread.get()

            Args:
                group_id (str): group_id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['group_id'] = \
                group_id
            return self.call_with_http_info(**kwargs)

        if self.delete_group is None:
            self.delete_group = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups/{group_id}',
                    'operation_id': 'delete_group',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'group_id',
                        'org_id',
                    ],
                    'required': [
                        'group_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'group_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'group_id': 'group_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'group_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_group
            )

        def __delete_group_member(
            self,
            group_id,
            member_id,
            **kwargs
        ):
            """Remove a group member  # noqa: E501

            Remove a group member  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_group_member(group_id, member_id, async_req=True)
            >>> result = thread.get()

            Args:
                group_id (str): group_id path
                member_id (str): member_id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['group_id'] = \
                group_id
            kwargs['member_id'] = \
                member_id
            return self.call_with_http_info(**kwargs)

        if self.delete_group_member is None:
            self.delete_group_member = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups/{group_id}/members/{member_id}',
                    'operation_id': 'delete_group_member',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'group_id',
                        'member_id',
                        'org_id',
                    ],
                    'required': [
                        'group_id',
                        'member_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'group_id':
                            (str,),
                        'member_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'group_id': 'group_id',
                        'member_id': 'member_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'group_id': 'path',
                        'member_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_group_member
            )

        def __get_group(
            self,
            group_id,
            **kwargs
        ):
            """Get a group  # noqa: E501

            Get a group  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_group(group_id, async_req=True)
            >>> result = thread.get()

            Args:
                group_id (str): group_id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                flatten_hierarchy (bool): Return the object with the full list of child members . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Group
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['group_id'] = \
                group_id
            return self.call_with_http_info(**kwargs)

        if self.get_group is None:
            self.get_group = _Endpoint(
                settings={
                    'response_type': (Group,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups/{group_id}',
                    'operation_id': 'get_group',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'group_id',
                        'org_id',
                        'flatten_hierarchy',
                    ],
                    'required': [
                        'group_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'group_id':
                            (str,),
                        'org_id':
                            (str,),
                        'flatten_hierarchy':
                            (bool,),
                    },
                    'attribute_map': {
                        'group_id': 'group_id',
                        'org_id': 'org_id',
                        'flatten_hierarchy': 'flatten_hierarchy',
                    },
                    'location_map': {
                        'group_id': 'path',
                        'org_id': 'query',
                        'flatten_hierarchy': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_group
            )

        def __list_groups(
            self,
            **kwargs
        ):
            """Get all groups  # noqa: E501

            Get all groups  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_groups(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                type ([str]): The type of groups to search for. Multiple values are ORed together.. [optional] if omitted the server will use the default value of ["group"]
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                previous_email (Email): Pagination based query with the user's email as the key. To get the initial entries supply an empty string.. [optional]
                search_direction (str): Direction which the search should go starting from the email_nullable_query parameter. . [optional] if omitted the server will use the default value of "forwards"
                prefix_email_search (Email): Keyword used to search for a list of users based on email. This parameter is case insensitive and finds users with an email that matches the keyword by its prefix. For example, if the keyword \"Foo\" is supplied to this parameter, users with emails of \"foo1@example.com\" and \"Foo2@test.com\" could be returned. . [optional]
                allow_partial_match (bool): Perform a case insensitive partial match of any string query parameters included in the query . [optional]
                first_name (str): query for users with a first name that matches the query parameter. [optional]
                last_name (str): query for users with a last name that matches the query parameter. [optional]
                search_params ([str]): A list of strings to perform a case-insensitive search on all relevant fields in the database for a given collection. Multiple values are ANDed together . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListGroupsResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_groups is None:
            self.list_groups = _Endpoint(
                settings={
                    'response_type': (ListGroupsResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups',
                    'operation_id': 'list_groups',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'org_id',
                        'type',
                        'limit',
                        'previous_email',
                        'search_direction',
                        'prefix_email_search',
                        'allow_partial_match',
                        'first_name',
                        'last_name',
                        'search_params',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                        'type',
                        'search_direction',
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                        ('type',): {

                            "GROUP": "group",
                            "SYSGROUP": "sysgroup",
                            "BIGROUP": "bigroup"
                        },
                        ('search_direction',): {

                            "BACKWARDS": "backwards",
                            "FORWARDS": "forwards"
                        },
                    },
                    'openapi_types': {
                        'org_id':
                            (str,),
                        'type':
                            ([str],),
                        'limit':
                            (int,),
                        'previous_email':
                            (Email,),
                        'search_direction':
                            (str,),
                        'prefix_email_search':
                            (Email,),
                        'allow_partial_match':
                            (bool,),
                        'first_name':
                            (str,),
                        'last_name':
                            (str,),
                        'search_params':
                            ([str],),
                    },
                    'attribute_map': {
                        'org_id': 'org_id',
                        'type': 'type',
                        'limit': 'limit',
                        'previous_email': 'previous_email',
                        'search_direction': 'search_direction',
                        'prefix_email_search': 'prefix_email_search',
                        'allow_partial_match': 'allow_partial_match',
                        'first_name': 'first_name',
                        'last_name': 'last_name',
                        'search_params': 'search_params',
                    },
                    'location_map': {
                        'org_id': 'query',
                        'type': 'query',
                        'limit': 'query',
                        'previous_email': 'query',
                        'search_direction': 'query',
                        'prefix_email_search': 'query',
                        'allow_partial_match': 'query',
                        'first_name': 'query',
                        'last_name': 'query',
                        'search_params': 'query',
                    },
                    'collection_format_map': {
                        'type': 'multi',
                        'search_params': 'multi',
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_groups
            )

        def __replace_group(
            self,
            group_id,
            **kwargs
        ):
            """update a group  # noqa: E501

            update a group  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_group(group_id, async_req=True)
            >>> result = thread.get()

            Args:
                group_id (str): group_id path

            Keyword Args:
                group (Group): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Group
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['group_id'] = \
                group_id
            return self.call_with_http_info(**kwargs)

        if self.replace_group is None:
            self.replace_group = _Endpoint(
                settings={
                    'response_type': (Group,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/groups/{group_id}',
                    'operation_id': 'replace_group',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'group_id',
                        'group',
                    ],
                    'required': [
                        'group_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'group_id':
                            (str,),
                        'group':
                            (Group,),
                    },
                    'attribute_map': {
                        'group_id': 'group_id',
                    },
                    'location_map': {
                        'group_id': 'path',
                        'group': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_group
            )

    add_group_member = None 
    create_group = None 
    create_upstream_group_reconcile = None 
    create_upstream_group_reconcile_sim = None 
    delete_group = None 
    delete_group_member = None 
    get_group = None 
    list_groups = None 
    replace_group = None 



import datetime
import json
import shutil
import urllib.parse

import agilicus
import requests

from . import context, hash, response
from .input_helpers import get_org_from_input_or_ctx, strip_none
from .input_helpers import update_org_from_input_or_ctx
from .output.table import column, spec_column, format_table, metadata_column, subtable

FILES_BASE_URI = "/v1/files"

OPER_STATUS_OPTIONS = ["active", "pending_delete", "deleted", "down"]


def query(ctx, org_id=None, tag=None, **kwargs):
    token = context.get_token(ctx)

    headers = {}
    headers["Authorization"] = "Bearer {}".format(token)

    params = {}

    if org_id:
        params["org_id"] = org_id
    else:
        org_id = context.get_org_id(ctx, token)
        if org_id:
            params["org_id"] = org_id

    if tag:
        params["tag"] = tag

    kwargs = strip_none(kwargs)
    params.update(**kwargs)
    query = urllib.parse.urlencode(params)
    uri = "{}?{}".format(FILES_BASE_URI, query)
    resp = requests.get(
        context.get_api(ctx) + uri,
        headers=headers,
        verify=context.get_cacert(ctx),
    )
    response.validate(resp)
    return json.loads(resp.text)["files"]


def upload(
    ctx,
    filename,
    region=None,
    org_id=None,
    tag=None,
    name=None,
    label=None,
    visibility=None,
    **kwargs,
):
    token = context.get_token(ctx)

    headers = {}
    headers["Authorization"] = "Bearer {}".format(token)

    if not name:
        name = filename

    if not label:
        label = datetime.datetime.utcnow().isoformat()

    md5_hash = hash.get_base64_md5(filename)

    multipart_form_data = {}
    multipart_form_data["md5_hash"] = md5_hash
    multipart_form_data["file_zip"] = (name, open(filename, "rb"))

    if org_id:
        multipart_form_data["org_id"] = (None, org_id)
    else:
        multipart_form_data["org_id"] = (None, context.get_org_id(ctx, token))

    if tag:
        multipart_form_data["tag"] = (None, tag)

    if region:
        multipart_form_data["region"] = (None, region)

    if name:
        multipart_form_data["name"] = (None, name)

    if label:
        multipart_form_data["label"] = (None, label)

    if visibility:
        multipart_form_data["visibility"] = (None, visibility)

    uri = "{}".format(FILES_BASE_URI)
    resp = requests.post(
        context.get_api(ctx) + uri,
        headers=headers,
        files=multipart_form_data,
        verify=context.get_cacert(ctx),
        timeout=60,
    )
    response.validate(resp)
    return json.loads(resp.text)


def reupload(
    ctx,
    file_id,
    filename,
    **kwargs,
):
    apiclient = context.get_apiclient_from_ctx(ctx)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    result = apiclient.files_api.reupload_file(
        file_id=file_id, file_zip=open(filename, "rb"), org_id=org_id
    )
    return result.to_dict()


def delete(ctx, file_id, org_id=None, _continue_on_error=False):
    token = context.get_token(ctx)

    headers = {}
    headers["Authorization"] = "Bearer {}".format(token)

    params = {}

    if org_id:
        params["org_id"] = org_id
    else:
        org_id = context.get_org_id(ctx, token)
        if org_id:
            params["org_id"] = org_id

    query = urllib.parse.urlencode(params)
    uri = "{}/{}?{}".format(FILES_BASE_URI, file_id, query)
    resp = requests.delete(
        context.get_api(ctx) + uri,
        headers=headers,
        verify=context.get_cacert(ctx),
    )
    response.validate(resp, _continue_on_error=_continue_on_error)
    return resp.text


def get(ctx, file_id, org_id=None):
    token = context.get_token(ctx)

    headers = {}
    headers["Authorization"] = "Bearer {}".format(token)

    params = {}

    if org_id:
        params["org_id"] = org_id
    else:
        org_id = context.get_org_id(ctx, token)
        if org_id:
            params["org_id"] = org_id

    query = urllib.parse.urlencode(params)
    uri = "{}/{}?{}".format(FILES_BASE_URI, file_id, query)
    resp = requests.get(
        context.get_api(ctx) + uri,
        headers=headers,
        verify=context.get_cacert(ctx),
    )
    response.validate(resp)
    return json.loads(resp.text)


def download(ctx, file_id, org_id=None, destination=None):
    token = context.get_token(ctx)

    headers = {}
    headers["Authorization"] = "Bearer {}".format(token)

    params = {}

    if org_id:
        params["org_id"] = org_id
    else:
        org_id = context.get_org_id(ctx, token)
        if org_id:
            params["org_id"] = org_id

    if not destination:
        _file = get(ctx, file_id, org_id)
        destination = _file["name"]

    query = urllib.parse.urlencode(params)
    uri = "{}_download/{}?{}".format(FILES_BASE_URI, file_id, query)
    with requests.get(
        context.get_api(ctx) + uri,
        headers=headers,
        verify=context.get_cacert(ctx),
        stream=True,
    ) as handle:
        handle.raise_for_status()
        with open(destination, "wb") as f:
            shutil.copyfileobj(handle.raw, f)


def list_file_associations(ctx, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    params = strip_none(kwargs)
    query_results = apiclient.files_api.list_file_associations(**params)
    return query_results.file_associations


def show_file_association(ctx, association_id, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    params = strip_none(kwargs)
    return apiclient.files_api.get_file_association(
        file_association_id=association_id, **params
    )


def delete_file_association(ctx, association_id, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    params = strip_none(kwargs)
    return apiclient.files_api.delete_file_association(
        file_association_id=association_id, **params
    )


def add_file_association(ctx, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    kwargs = strip_none(kwargs)
    spec = agilicus.FileAssociationSpec(**kwargs)
    assoc = agilicus.FileAssociation(spec=spec)
    return apiclient.files_api.create_file_association(assoc)


def clear_file_associations(ctx, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    params = strip_none(kwargs)
    req = agilicus.ClearFileAssociationRequest(**params)
    return apiclient.files_api.create_file_association_clear_task(req)


_ASSOC_COLUMNS = [
    metadata_column("id"),
    spec_column("org_id"),
    spec_column("file_id"),
    spec_column("object_id"),
]


def format_file_associations(ctx, file_associations):
    columns = _ASSOC_COLUMNS
    return format_table(ctx, file_associations, columns)


def format_cleared_associations(ctx, cleared_associations):
    file_columns = [
        column("id"),
        column("name"),
        column("tag"),
    ]
    columns = [
        subtable(ctx, "deleted_associations", _ASSOC_COLUMNS),
        subtable(ctx, "pending_delete_files", file_columns),
    ]

    return format_table(ctx, cleared_associations, columns)



import click

from . import context
from .input_helpers import get_org_from_input_or_ctx, update_if_not_none
from .input_helpers import get_user_id_from_input_or_ctx
from .input_helpers import update_org_from_input_or_ctx, strip_none
from agilicus.output.table import make_columns


from .output.table import (
    subobject_column,
    format_table,
    metadata_column,
    spec_column,
    status_column,
    subtable,
)
import agilicus

permissioned_resource_types = [
    "application",
    "fileshare",
    "application_service",
    "desktop",
    "group",
    "launcher",
    "ssh",
]
resource_types = permissioned_resource_types + ["service_forwarder"]
permissioned_resource_type_enum = click.Choice(permissioned_resource_types)
resource_type_enum = click.Choice(resource_types)


def query_permissions(ctx, org_id=None, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    query_results = apiclient.permissions_api.list_resource_permissions(
        org_id=org_id, **strip_none(kwargs)
    )
    if query_results:
        return query_results.resource_permissions
    return []


def render_permissions(ctx, org_id=None, user_id=None, limit=None, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    user_id = get_user_id_from_input_or_ctx(ctx, user_id=user_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    query_results = apiclient.user_api.list_all_resource_permissions(
        org_id=org_id, user_id=user_id, **strip_none(kwargs)
    )
    if query_results:
        return query_results.resource_permissions
    return []


def format_permissions(ctx, roles):
    columns = [
        metadata_column("id"),
        spec_column("resource_type", "type"),
        spec_column("user_id", "user id"),
        spec_column("org_id", "org id"),
        spec_column("resource_id", "resource id"),
        spec_column("resource_role_name", "role"),
    ]
    return format_table(ctx, roles, columns)


def add_permission(
    ctx, user_id, resource_id, resource_type, resource_role_name, org_id=None
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    spec = agilicus.ResourcePermissionSpec(
        org_id=org_id,
        user_id=user_id,
        resource_type=resource_type,
        resource_role_name=resource_role_name,
        resource_id=resource_id,
    )
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    return apiclient.permissions_api.create_resource_permission(
        agilicus.ResourcePermission(spec=spec)
    ).to_dict()


def delete_permission(ctx, id, org_id=None):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    return apiclient.permissions_api.delete_resource_permission(id, org_id=org_id)


def bulk_delete_permission(ctx, **kwargs):
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    apiclient = context.get_apiclient_from_ctx(ctx)
    return apiclient.permissions_api.bulk_delete_resource_permission(
        **strip_none(kwargs)
    )


def query_resources(ctx, org_id=None, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs["org_id"] = org_id
    params = {}
    update_if_not_none(params, kwargs)
    query_results = apiclient.resources_api.list_resources(**params)
    if query_results:
        return query_results.resources
    return []


def format_resources(ctx, resources, show_columns=None, reset_columns=None):
    columns = make_columns(
        ctx,
        resources,
        """
          - metadata.id(newname=id)
          - spec.resource_type(newname=resource_type)
          - spec.name(newname=name)
          - status.resource_stats.overall_status(newname=status)
        """,
        show=show_columns,
        clear=reset_columns,
    )
    return format_table(ctx, resources, columns)


def query_roles(ctx, org_id=None, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    query_results = apiclient.permissions_api.list_resource_roles(
        org_id=org_id, **strip_none(kwargs)
    )
    return query_results.resource_roles


def format_roles(ctx, roles):
    columns = [
        spec_column("resource_type", "type"),
        spec_column("role_name", "name"),
    ]
    return format_table(ctx, roles, columns)


def create_resource_group(ctx, resource_members, name, org_id=None):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    member_objs = []
    for member in resource_members:
        member_objs.append(agilicus.ResourceMember(id=member))
    spec = agilicus.ResourceSpec(
        resource_members=member_objs, resource_type="group", name=name, org_id=org_id
    )
    resource = agilicus.Resource(spec=spec)

    return apiclient.resources_api.add_resource(resource).to_dict()


def delete_resource(ctx, id, org_id=None):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    return apiclient.resources_api.delete_resource(id, org_id=org_id)


def _get_resource(ctx, id, org_id):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    return apiclient.resources_api.get_resource(id, org_id=org_id)


def get_resource(ctx, id, org_id=None, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    resource = _get_resource(ctx, id, org_id)
    return resource.to_dict()


def update_resource(
    ctx,
    id,
    resource_members=None,
    remove_resource_members=None,
    name=None,
    org_id=None,
    rules_config_file=None,
    published=None,
    **kwargs,
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    resource = apiclient.resources_api.get_resource(id, org_id=org_id)
    if remove_resource_members is not None:
        old_members = resource.spec.resource_members
        resource.spec.resource_members = []
        for member in old_members:
            if member.id in remove_resource_members:
                # needs to be removed.
                continue
            resource.spec.resource_members.append(member)

    if resource_members is not None:
        for member in resource_members:
            resource.spec.resource_members.append(agilicus.ResourceMember(id=member))
    if name is not None:
        resource.spec.name = name
    if rules_config_file is not None:
        resource.spec.config.rules_config = rules_config_file
    if published is not None:
        resource.spec.config.published = published

    return apiclient.resources_api.replace_resource(id, resource=resource).to_dict()


def add_display_info_icon(
    ctx, resource_id, uri, purpose, height_px, width_px, org_id, **kwargs
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    resource = _get_resource(ctx, resource_id, org_id=org_id)

    config = resource.spec.config
    if config is None:
        config = agilicus.ResourceConfig()
        resource.spec.config = config

    display_info = config.display_info
    if display_info is None:
        display_info = agilicus.DisplayInfo(icons=[])
        config.display_info = display_info

    purposes = [agilicus.IconPurpose(p) for p in purpose]
    new_icon = agilicus.Icon(uri=uri, purposes=purposes)
    if height_px is not None and width_px is not None:
        new_icon.dimensions = agilicus.IconDimensions(width=width_px, height=height_px)
    display_info.icons.append(new_icon)

    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    return apiclient.resources_api.replace_resource(
        resource_id, resource=resource
    ).to_dict()


def delete_display_info_icon(ctx, resource_id, uri, org_id, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    resource = _get_resource(ctx, resource_id, org_id=org_id)

    config = resource.spec.config
    if not config:
        return resource.to_dict()

    display_info = config.display_info
    if not display_info:
        return resource.to_dict()

    icons = [icon for icon in display_info.icons if icon.uri != uri]
    display_info.icons = icons

    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    return apiclient.resources_api.replace_resource(
        resource_id, resource=resource
    ).to_dict()


def list_combined_resource_rules(ctx, org_id=None, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id, **kwargs)
    kwargs = strip_none(kwargs)
    query_results = apiclient.resources_api.list_combined_resource_rules(
        org_id=org_id, **kwargs
    )
    if query_results:
        return query_results.combined_resource_rules
    return []


def format_combined_resource_rules_as_text(ctx, resource_rules):
    def node_column(name):
        return subobject_column(name, name, "node")

    rules_columns = [
        subobject_column("priority", "priority", "node"),
        subobject_column("rule.name", "name", "node"),
    ]

    columns = [
        status_column("org_id"),
        status_column("resource_id"),
        status_column("scope"),
        status_column("role_name"),
        subtable(ctx, "rules", rules_columns, subobject_name="status"),
    ]

    return format_table(ctx, resource_rules, columns)
